-- ALU_16.vhd
-- version 20080102
-- This file gathers the separate units
-- and connects them so that latency is minimal.

--  _EN         _ENr                      _ENrr
-- ASU :  ---FF------------------------FFx---------------L8----\MUX0___Result
-- SR :   ---FF--[......extern......]--FF1-\_MUX2__m2__MUX1_m1_/(L9)
-- ROP2 : ---FF------------L5-\_MUX4___FF2_/ (L7)      (L8)
-- SHL  : ---FF------------L5-/ (L6)                   /
-- IE     ---FF---\MUX5-m5-IE---L2-----FF3-\_MUX3_m3__/
-- Mem     IOFF---/------------------------/

--  IP, NIP, Mode ?


library ieee;
use ieee.std_logic_1164.all;

entity ALU_16 is
  generic (
    YASEP_SIZE : integer := 16;
    PIPELINED  : integer := 1   -- pipelined when 1
  );
  port(
    DataA, DataB, MemRead, SR_Read : in std_logic_vector(YASEP_SIZE-1 downto 0);
    clk, SR_EN, ROP2_EN, MEM_ENr, MEM_ENrr,
     add_EN, Addsub, ROP2_S1, ROP2_S2, ROP2_X2,
     SHL_en, Rotation, Direction, Arithmetic,
     IE_en, Load, SignExt, OffsetA : in std_logic;

    Result : out std_logic_vector(YASEP_SIZE-1 downto 0);
    Carry_out, ROP2_Combine : out std_logic);
end ALU_16;

architecture together of ALU_16 is
  signal ROP2_OUT, ResSHL, ResIE, FF1, FF2, FF3, DataAr, DataBr,
          m1, m2, m3, m5 : std_logic_vector(YASEP_SIZE-1 downto 0);
  signal Sum : std_logic_vector(YASEP_SIZE downto 0);
  signal IE_enr,  SR_ENr,  add_ENr,  SHL_enr, ROP2_ENr,
         IE_enrr, SR_ENrr, add_ENrr,
         Loadr, SignExtr, OffsetAr : std_logic;
begin


ASU:
  entity work.ASUROP2_16(ActelProASIC3)
  generic map(
  	PIPELINED  => PIPELINED,
    YASEP_SIZE => YASEP_SIZE
  )
  port map(
    DataA  => DataB,
    DataB  => DataA, -- DataB and DataA are a bit mixed up :-/
    Addsub => Addsub,
    clk    => clk,
    add_EN => add_EN,
    ROP2_S1 => ROP2_S1,
    ROP2_S2 => ROP2_S2,
    ROP2_X2 => ROP2_X2,

    ROP2_OUT => ROP2_OUT,
    ROP2_combine => ROP2_Combine, -- direct output, i'll hack it later.
    Sum    => Sum
  );

SHL:
  entity work.SHL_16(behaviour)
  generic map(
  	PIPELINED  => PIPELINED,
    YASEP_SIZE => YASEP_SIZE
  )
  port map(
    DataA      => DataA,
    SHAmount   => DataB(3 downto 0),
    clk        => clk,
    SHL_en     => SHL_en,
    Rotation   => Rotation,
    Direction  => Direction,
    Arithmetic => Arithmetic,

    ResultSHL  => ResSHL
  );

IE:
    entity work.IE_16(second)
    generic map(
      YASEP_SIZE => YASEP_SIZE
    )
    port map(
      DataA    => m5,
      DataB    => DataBr,
      Load     => Loadr,
      SignExt  => SignExtr,
      OffsetA  => OffsetAr,

      ResultIE => ResIE
    );


-- signal delays :
  process(clk, add_ENr, IE_enr, SR_ENr,
     Load, SignExt, OffsetA, IE_en, SR_EN, add_EN, SHL_en, ROP2_EN)
  begin
    if (clk'event and clk='1') then
 -- 2nd stage
      add_ENrr <= add_ENr;
      IE_enrr  <= IE_enr;
      SR_ENrr  <= SR_ENr;

 -- 1st stage
      Loadr    <= Load;
      SignExtr <= SignExt;
      OffsetAr <= OffsetA;
      IE_enr   <= IE_en;
      SR_ENr   <= SR_EN;
      add_ENr  <= add_EN;
      SHL_enr  <= SHL_en;
      ROP2_ENr <= ROP2_EN;
      DataAr   <= DataA;
      DataBr   <= DataB;
    end if;
  end process;


-- pipeline gates

  process(clk, SR_Read, SR_EN, ROP2_OUT, ResSHL, ROP2_ENr, SHL_enr, ResIE, IE_enr)
  begin
    if (PIPELINED=0) or (clk'event and clk='1' and SR_EN='1') then
      FF1 <= SR_Read;
    end if;

    if (PIPELINED=0) or (clk'event and clk='1' and (ROP2_ENr or SHL_enr)='1') then
      -- MUX4
      if (SHL_enr = '1') then
        FF2 <= ResSHL;
      else
        FF2 <= ROP2_OUT;
      end if;
    end if;

    if (PIPELINED=0) or (clk'event and clk='1' and IE_enr='1') then
      FF3 <= ResIE;
    end if;

  end process;


-- MUX :

  m5 <= MemRead when (MEM_enr = '1')
   else DataAr;

  m3 <= FF3 when (IE_enrr = '1')
   else MemRead;

  -- MUX2
  m2 <= FF1 when (SR_ENrr ='1')
   else FF2;

  -- MUX1
  m1 <= m3 when ((MEM_ENrr or IE_enrr)='1')
   else m2;

  -- MUX0
  Result <= Sum(YASEP_SIZE-1 downto 0) when (add_ENrr = '1')
   else m1;

  Carry_out <= Sum(YASEP_SIZE);
end together;
