-- IE_16.vhd
-- Version 20090102 by whygee@f-cpu.org
-- This code implements the Insert/Extract unit of YASEP-16
-- The opcodes are LSB, LZB, SB and MOV(A/B)
-- LZB can also be used to extend the sign of a byte in a register.
-- This version removes the pipeline gates.

-- port A  : src/dest  => original value of the register, to be mixed with B
-- OffsetA : LSB of the pointer associated to an eventual
-- port B  : src/Imm16 => value to be (partially) assigned for MOV or LxB

-- opcode :        MOV(B) MOV(A)   SB   LZB    LSB
-- Load              0      0       0    1      1
-- SignExt           0      0       1    0      1
-- pointer LSB :     0      1      0/1  0/1    0/1


library IEEE;
use ieee.std_logic_1164.all;

entity IE_16 is
  generic (
    YASEP_SIZE : integer := 16 -- either 16 or 32
  );
  port(
    DataA, DataB : in std_logic_vector(YASEP_SIZE-1 downto 0);
    Load, SignExt, OffsetA : in std_logic;
    ResultIE : out std_logic_vector(YASEP_SIZE-1 downto 0)
  );
end IE_16;

-------------------------------------------------------------------
architecture second of IE_16 is
begin

  process (DataA, DataB, Load, SignExt, OffsetA)
    variable DataAr : std_logic_vector(DataA'range);
    variable loS, hiS, loRes, hiRes : std_logic_vector(7 downto 0);
    variable SB0, SB1, Ext : std_logic;

    begin

-- opcode :        MOV(B)      SB           LZB                LSB
-- Load              0          0            1                  1
-- SignExt           0          1            0                  1
-- OffsetA :         X       0     1      0     1        0            1
-- result Lo       lo(B)   lo(B) lo(A)  lo(B)  hi(B)   lo(B)        hi(B)
-- result Hi       hi(B)   hi(A) lo(B)    0     0    sign(lo(B))  sign(hi(B))

-- SB/LxB can be factored with a preliminary byte swap :
      if (OffsetA = '0') then
        loS := DataB( 7 downto 0);
        hiS := DataB(15 downto 8);
      else
        loS := DataB(15 downto 8);
        hiS := DataB( 7 downto 0);
      end if;

-- after the swap, the table is much simpler :
-- opcode :        MOV(B) SWP(B)       SB           LZB                LSB
-- pointer LSB :     0      1       0     1      0     1         0            1
-- result Lo       lo(S)  lo(S)   lo(S) lo(A)  lo(S)  lo(S)    lo(S)        lo(S)
-- result Hi       hi(S)  hi(S)   hi(A) hi(S)    0     0     sign(lo(S))  sign(lo(S))
-- byteswap comes for free, a corner case of MOV :-)

-- however, because the memory read port comes here,
-- it is desirable to have a MOV(A) (a bypass) :
-- opcode :        MOV(B) MOV(A)      SB           LZB                LSB
-- Load              0      0       0     0      1     1         1            1
-- SignExt           0      0       1     1      0     0         1            1
-- pointer LSB :     0      1       0     1      0     1         0            1
-- result Lo       lo(S)  lo(A)   lo(S) lo(A)  lo(S)  lo(S)    lo(S)        lo(S)
-- result Hi       hi(S)  hi(A)   hi(A) hi(S)    0     0     sign(lo(S))  sign(lo(S))
-- ==>   MOVA := not(Load) and not(SignExt) and OffsetA;

-- for the Lo result, we see that it's always lo(S)
-- except for SB with LSB=1, or MOVA, they both return lo(A)
-- so we precompute this condition :
    SB1 := not(Load) and OffsetA;

-- for the Hi result, it's either hi(S)
-- (except for SB with LSB=0 OR MOVA)
    SB0 := not(Load) and (SignExt xor OffsetA);
-- or it's the sign-extended value of lo(S)
    Ext := loS(7) and SignExt;

-- The result :
      -- MUX the lower byte :
      if (SB1 = '1') then
        loRes := DataA(7 downto 0);
      else
        loRes := loS; -- all the other cases
      end if;

      -- MUX the higher byte :
      if (Load = '1') then
        hiRes := (hiRes'range => Ext);  -- Load signed or unsigned
      else
        if (SB0 = '1') then
          hiRes := DataA(15 downto 8);
        else
          hiRes := hiS;  
        end if;
      end if;

      -- concatenate the result
      ResultIE <= hiRes & loRes;

    end process;
end second;
