-- ASU_ROP2.vhd
-- version 200808019 : add_EN
-- 32-bit, pipelined : reached 116MHz
-- 32-bit, not pipelined : about 70MHz (incl. a final MUX)

-- This unit performs the addition, substraction and all logic operations ("ROP2")
-- for the YASEP in 16 and 32-bit wide datapaths.
-- It is designed to operate > 100MHz in pipelined mode
-- on "standard"-grade Actel ProASIC3 chips
-- (hence the 3-input logic operations).
-- On faster a FPGA or in ASIC, the "behaviour"
-- implementation should be fast enough, but the
-- "detailed" version is handy anyway, if one needs to reach
-- "somewhat faster than out-of-the-box" speeds.

-- about the coding style : it's awful, i know.
-- but i spent half of the time fighting the tools,
-- the other half fighting the langage, and the
-- remaining time implementing useful features.

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity AddSub32 is
  generic (
  	PIPELINED  : integer := 1; -- 1=pipelined
    YASEP_SIZE : integer := 32 -- either 16 or 32
  );
  port(
    DataA, DataB : in std_logic_vector(YASEP_SIZE-1 downto 0);
    Addsub, clk, add_en : in std_logic;
    Sum : out std_logic_vector(YASEP_SIZE downto 0);

    ROP2_S1, ROP2_S2, ROP2_X2 : in std_logic; -- ROP2 functions
    ROP2_OUT : out std_logic_vector(YASEP_SIZE-1 downto 0);
    ROP2_combine : out std_logic);
end AddSub32;

-----------------------------------------------------------------

architecture behaviour of AddSub32 is
begin

  -- all the unit's inputs, in case the thing is not pipelined :
  process (DataA, DataB, Addsub, clk, add_en, ROP2_S1, ROP2_S2, ROP2_X2 )

    variable sumAux : unsigned(Sum'range);
    variable DataAReg, DataBReg, DataBReg2,
         ROP2_RS1, ROP2_RS2, ROP2_RX2 : std_logic_vector(DataA'range);
    variable AddSub2, ROP2_c, ROP2_S1r, ROP2_S2r, ROP2_X2r : std_logic;

  begin

    assert (YASEP_SIZE=16) or (YASEP_SIZE=32)
      report "YASEP_SIZE must be either 16 or 32"
      severity failure;

    -- input pipeline gates :    
    if (PIPELINED=0) or (clk'event and clk='1' and add_en='1') then
      DataAReg := DataA;
      DataBReg := DataB;
      Addsub2  := Addsub;
      ROP2_S1r := ROP2_S1;
      ROP2_S2r := ROP2_S2;
      ROP2_X2r := ROP2_X2;
    end if;

    -- add/sub :
    if (Addsub2 = '0') then
      sumAux := unsigned('0' & DataAReg) + unsigned('0' & DataBReg);
    else
      sumAux := unsigned('0' & DataAReg) - unsigned('0' & DataBReg);
    end if;

    -- ROP2 :
    -- Beware : the "ActelProASIC3" architecture version reuses the first
    --    logic stages of the add/sub unit and the XOR of DataB is done 1 step before
    --     so operating frequencies might vary
    DataBReg2 := DataB xor (DataB'range=>Addsub); -- initial XOR

    if (ROP2_S1r='1') then -- First MUX
      ROP2_RS1 := DataAReg and DataBReg2;
    else
      ROP2_RS1 := DataAReg xor DataBReg2;
    end if;

    if (ROP2_S2r='1') then -- second mux
      ROP2_RS2 := DataAReg  or DataBReg2;
    else
      ROP2_RS2 := ROP2_RS1;
    end if;

    ROP2_RX2 := ROP2_RS2 xor (ROP2_RX2'range => ROP2_X2r); -- last XOR

    -- OR of all the XORs (reduction)
    -- The source is not "DataAReg xor DataBReg2" but ROP2_RS1
    -- so that the fanout of the XOR is not excessive.
    -- this may be pointless with a LUT-4 architecture, though.
    if ROP2_RS1=(ROP2_RS1'range=>'0') then
      ROP2_c := '0';
    else
      ROP2_c := '1';
    end if;

    -- output pipeline gates :
    if (PIPELINED=0) or (clk'event and clk='1') then --  and add_EN2='1'
      Sum(Sum'left)            <= std_logic(       sumAux(Sum'left) xor Addsub2);
      Sum(Sum'left-1 downto 0) <= std_logic_vector(sumAux(Sum'left-1 downto 0));
      ROP2_OUT <= ROP2_RX2;
      ROP2_combine <= ROP2_c;
    end if;

--    if (clk'event and clk='1') then
--      add_EN2 := add_en;
--    end if;

  end process;
end behaviour;


-----------------------------------------------------------------

architecture ActelProASIC3 of AddSub32 is
begin

  -- all the unit's inputs again :-/
  process (DataA, DataB, Addsub, clk, add_en, ROP2_S1, ROP2_S2, ROP2_X2 )

  -- when in 16-bit mode, this remaps the high half-word indices to the lower,
  -- to keep modelsim happy. The higher half is not accessed because
  -- of the IFs so the design works (but it's painfully ugly)
  variable dummy_base : integer := 32 - YASEP_SIZE;

  variable DataAReg, DataBReg,
         uDataB_s1in, Gene, Prop, Comp, -- preliminary inversions
         ROP2_RS1, ROP2_RS2, ROP2_RX2 : -- ROP2 temporary results
         std_logic_vector(YASEP_SIZE-1 downto 0);
  variable DataB0, Sum32_or, N391r, Sum0r, Sum0rr, Sum1rr, Sum2rr,
          Prop1pre, Prop1preR, Gene1pre, Gene1preR,
          N555_dup, N621_dup, N619_dup, N594_dup, N662_dup, -- optimisation CDP
          ROP2_c, -- add_EN2,
-- latches level 2 :
--L2
          Comp5r, Comp7r, Comp8r, Prop8r, Gene8r, Comp9r, Comp10r, Comp11r, Comp12r, Gene12r,
          Prop12r, Comp13r, Comp14r, Comp15r, Comp16r, Comp17r, Comp18r, Comp19r, Comp20r, Comp21r,
          Comp22r, Comp23r, Comp24r, Comp25r, Comp26r, Comp28r, Comp29r, Comp30r, Comp31r,
--L3
          N109_1r, N105_1r, N524r, N516r,
--L4
          N611r, I300Y1r,
--L5
          I271un1Y0r, I270Y0a01r, I358YN4ir, N4ir, N562r, N630r, N632r, N634r,
          N638r, N642r, N646r, N647r, N650r, N657r, N658r, N662r, N663r,
          N678r, N662_dupr, I208un1Yr, I268Y2r,
-- L6
          Sum3r, Sum4r, I210Y0a3r, Sum6r, N926r, N920r, N917r, N914r,
          N908r, I250un1Yir, N731r, N732r, I244un1Yr, I281un1Y0r, I280un1Yr, I280Y0r,
          I279Y0r, I240un1Yr, I277un1Yr, I277Y0r, I276un1Y0r, I276Y0r, I275Y0r, I273Y1r,
          I272un1Y0r, I272Y1r, I271Y1r, I270Y2r, I269Y2r, I268un1Y0r, I300un1Y0r, I300Y3r, -- 85 latches

  I94un1Y, N391, -- 1
  N621, N554, N555, N555_Ain, N555_Aout, -- 2
  N619, -- 3
  N685, N617, N618, N550, N551, -- 4
  I210Y0a3, -- 5
  N681, N613, N614, N546, N547, -- 6
  I208un1Y, N611, N612, N_42, N548i, I146Y0, -- 7
  N926, N678, N677, N610, N609, N543, N542, -- 8
  I256Y0a3, -- 9
  N920, N674, N673, N605, N606, N538, N539, -- 10
  N917, N672, N671, N604, I139Y0, N_40, N603, N540i, I138Y0, -- 11
  N914, N602, N601, N669, N670, N535, N534, -- 12
  N911i, -- 13
  N908, N665, N666, N598, N531, N597, N530, -- 14
  N663, I250un1Yi, N595, N596, N38i0, N532, I130Y0, N664, -- 15
  N526, N527, N661, N662, N594, N593, N731, N732, -- 16
  I282Y0a, N524, N105_1, -- 17
  I244un1Y, I281Y0, I281un1Y0, N522, N523, N590, N657, N658, N589, -- 18
  I280un1Y, I280Y0, N_36, N588, N656, I122Y0, N587, N655, -- 19
  I279Y0, I240un1Y, N586, N519, N653, N518, N585, I279un1Y0, -- 20
  I278Y0a3, N516, N109_1, -- 21
  I277un1Y, I277Y0, N650, N582, N515, N649, N581, N514, -- 22
  I276un1Y, I276Y0, I276un1Y0, N648, N579, N580, N89_1, N647, N512, -- 23
  I275un1Y, I275Y0, N645, N646, N577, N578, N511, N510, -- 24
  N4i, I358YN4i, rm1, I274Yia300, N79, I358Ym3e0, I274Yio30, N508, dm1e, -- 25
  I273un1Y, N642, N506, N507, I273Y1, I273Y0, N573, N574, -- 26
  N869i, I272un1Y0, I272Y0, I272Y1, N640, N571, N572, N87_1, N504, -- 27
  N866i, I271un1Y0, I271Y1, I271Y0, N638, N569, N570, N502, N503, -- 28
  I270Y0a0, I270Y0a01, I270Y0a10, I270Y0a200, I270Y0, I270Y1, I270Y2, N78, N500, -- 29
  I269un1Y, I269Y0, I269Y1, I269Y2, N634, N566, N498, N499, -- 30
  I268un1Y, I268Y0, I268Y1, I268Y2, I268Y3, I268un1Y0, N632, N564, N497, -- 31
  I266un1Y, I300un1Y0, I300Y0, I300Y1, I300Y3, I300Y4, N495, N562, N630 : std_logic; -- carry

-- 20080811 : 116MHz pipelined, 32 bits
--  84MHz not pipelined, 32 bits (with final mux)

begin

    assert (YASEP_SIZE=16) or (YASEP_SIZE=32)
      report "YASEP_SIZE must be either 16 or 32"
      severity failure;

-- L0 : DataA, DataB, Addsub

------------------------------------------------------------
------------------------------------------------------------
-- L1

-- inversions prliminaires L1(0,0)
  uDataB_s1in  := DataB xor (DataB'range => Addsub);

--  I333_Y fo1 L1(0,0)
  Sum0r := DataA(0) xor DataB(0);

-- dcritisation d'un chemin :
   Prop1preR := (DataB(1) xor Addsub) or  DataA(1);
   Gene1preR := (DataB(1) xor Addsub) and DataA(1);

-- Sum1
  -- I0_G0N fo2 L1(0,0)
    N391r := DataA(0) and Addsub;

-- Sum2
    N555_Ain := DataA(0) or Addsub;

------------------------------------------------------------
------------------------------------------------------------
-- pipeline gate #1:
  if (PIPELINED=0) or (clk'event and clk='1' and add_en='1') then
    DataAReg  := DataA; -- 19 bits used in 32-bit mode so warnings about pruned gates are ok
    DataBReg  := uDataB_s1in;
    DataB0    := DataB(0);
    N555_Aout := N555_Ain;
    N391      := N391r;
    Sum0rr    := Sum0r;
    Prop1pre  := Prop1preR;
    Gene1pre  := Gene1preR;
  end if;

------------------------------------------------------------
------------------------------------------------------------
-- L2

-- Generate, Propagate, Compute : L2(1,0)
  Gene := DataBReg and DataA;
  Prop := DataBReg  or DataA;
  Comp := DataBReg xor DataA;

-- Sum1
  -- I94_un1_Y fo1 L2(0,0,1)
    I94un1Y := (DataAReg(0) or not(DataB0)) and DataBReg(0);

-- Sum2
  -- I92_Y fo2 L2(1,1,1)
    N554 := (N391 and Prop1pre) or Gene1pre;
  -- I93_Y fo3 L2(1,1)
    N555 := N555_Aout and Prop1pre;
  -- I93_Y fo1 L2(1,1)
    N555_dup := N555_Aout and Prop1pre and DataBReg(0);

------------------------------------------------------------
------------------------------------------------------------
-- L3

-- Sum1
  --  I334_Y L3(2,1,2)
    Sum1rr := (I94un1Y or N391) xor Comp(1);

-- Sum2
  -- I155_Y fo3 L3(2,1,2) 
    N621 := (N555 and DataBReg(0)) or N554;
  -- I155_Y fo2 L3(2,2)
    N621_dup := N555_dup or N554;

-- Sum4
  -- I88_Y fo2 L3(1,0,2)
    N550 := (DataBReg(3) and DataAReg(3))
         or (DataBReg(3) and Gene(2))
         or (DataAReg(3) and Gene(2));
  -- I89_Y fo3 L3(2,2)
    N551 := Prop(3) and Prop(2);

-- Sum6
  -- I84_Y fo2 L3(1,0,2)
    N546 := (DataBReg(5) and DataAReg(5))
         or (DataBReg(5) and Gene(4))
         or (DataAReg(5) and Gene(4));
  -- AND2 I85_Y fo3 L3(2,2)
    N547 := Prop(5) and Prop(4);

-- Sum7
  -- I83_Y_i fo2 L3(2,2)
    N_42 := Prop(5) and Prop(6);
  -- I86_Y_0 fo1 L3(2,2,2)
    N548i := (Gene(3) and Prop(4)) or Gene(4);
  -- I146_Y_0 fo1 L3(1,0,2)
    I146Y0 := (DataBReg(6) and DataAReg(6))
           or (DataBReg(6) and Gene(5))
           or (DataAReg(6) and Gene(5));

-- Sum8
  -- I80_Y fo2 L3(1,0,2)
    N542 := (DataBReg(7) and DataAReg(7))
         or (DataBReg(7) and Gene(6))
         or (DataAReg(7) and Gene(6));
  --  I81_Y fo3 L3(2,2)
    N543 := Prop(7) and Prop(6);

-- Sum10
  -- I76_Y fo2 L3(1,0,2)
    N538 := (DataBReg(9) and DataAReg(9))
         or (DataBReg(9) and Gene(8))
         or (DataAReg(9) and Gene(8));
  -- I77_Y fo3 L3(2,2)
    N539 := Prop(9) and Prop(8);

-- Sum11
  -- I138_Y_0 fo1 L3(2,2,2)
    I138Y0 := (Gene(9) and Prop(10)) or Gene(10);
  -- I139_Y_0 fo1 L3(2,2)
    I139Y0 := Prop(7) and Prop(8);
  -- I75_Y_i fo2 L3(2,2)
    N_40 := Prop(9) and Prop(10);
  -- I78_Y_0 fo1 L3(2,2,2)
    N540i := (Gene(7) and Prop(8)) or Gene(8);

-- Sum12
  -- I72_Y fo2 L3(2,0,1)
    N534 := (DataBReg(11) and DataAReg(11))
         or (DataBReg(11) and Gene(10))
         or (DataAReg(11) and Gene(10));
  -- I73_Y fo3 L3(2,2)
    N535 := Prop(11) and Prop(10);

-- Sum14
  -- I68_Y fo2 L3(2,1,0)
    N530 := (DataBReg(13) and DataAReg(13))
         or (DataBReg(13) and Gene(12))
         or (DataAReg(13) and Gene(12));
  -- I69 fo3 L3(2,2)
    N531 := Prop(13) and Prop(12);

-- Sum15
  -- I70_Y_0 fo1 L3(2,2,2)
    N532 := (Gene(11) and Prop(12)) or Gene(12);
  -- I67_Y_i fo2 L3(2,2)
    N38i0 := Prop(14) and Prop(13);
  -- I130_Y_0 fo1 L3(2,1,0)
    I130Y0 := (DataBReg(14) and DataAReg(14))
           or (DataBReg(14) and Gene(13))
           or (DataAReg(14) and Gene(13));

-- Sum16
  -- I64_Y fo2 L3(1,0,2)
    N526 := (DataBReg(15) and DataAReg(15))
         or (DataBReg(15) and Gene(14))
         or (DataAReg(15) and Gene(14));
  -- I65_Y fo3  L3(2,2)
    N527 := Prop(15) and Prop(14);

if YASEP_SIZE=32 then

-- Sum17
  -- I282_Y_0_a3_1 fo2 L3(2,2)
    N105_1 := Prop(15) and Prop(16);
  -- I62_0_o2 fo2 L3(2,2,2)
    N524 := (Gene(15) and Prop(16)) or Gene(16);

-- Sum18
  -- I60_Y fo2 L3(1,0,2)
    N522 := (DataBReg(17) and DataAReg(17))
         or (DataBReg(17) and Gene(16))
         or (DataAReg(17) and Gene(16));
  -- I61_Y fo3 L3(2,2)
    N523 := Prop(17) and Prop(16);

-- Sum19
  -- I122_Y_0 fo1 L3(2,2,2)
    I122Y0 := (DataBReg(18) and DataAReg(18))
           or (DataBReg(18) and Gene(17))
           or (DataAReg(18) and Gene(17));
  -- I59_Y_i fo2 L3(2,2)
    N_36 := Prop(18) and Prop(17);

-- Sum20
  -- I56_Y fo2 L3(2,2,2)
    N518 := (Gene(18) and Prop(19)) or Gene(19);
  -- I57_Y fo3 L3(2,2)
    N519 := Prop(19) and Prop(18);

-- Sum21
  -- I278_Y_0_a3_1 fo4 L3(2,2)
    N109_1 := Prop(20) and Prop(19);
  -- I54_Y_0_o2 fo5 L3(2,2,2)
    N516 := (Gene(19) and Prop(20)) or Gene(20);

-- Sum22
  -- I52_Y fo2 L3(2,1,0)
    N514 := (DataBReg(21) and DataAReg(21))
         or (DataBReg(21) and Gene(20))
         or (DataAReg(21) and Gene(20));
  -- I53_Y fo3 L3(2,2)
    N515 := Prop(21) and Prop(20);

-- Sum23
  -- I274_Y_i_a3_0_1 fo4 L3(2,2)
    N89_1 := Prop(21) and Prop(22);
  -- I50_Y_0_o2 fo2 L3(2,2,3)
    N512 := (Gene(21) and Prop(22)) or Gene(22);

-- Sum24
  -- I48_Y fo2 L3(2,0,1)
    N510 := (DataBReg(23) and DataAReg(23))
         or (DataBReg(23) and Gene(22))
         or (DataAReg(23) and Gene(22));
  -- I49_Y fo3 L3(2,2)
    N511 := Prop(23) and Prop(22);

-- Sum25
  -- I47_Y_i_o2 fo4 L3(2,2)
    N79 := Prop(23) and Prop(24);
  -- I46_Y_0_o2 fo2 L3(2,2,2)
    N508 := (Gene(23) and Prop(24)) or Gene(24);

-- Sum26
  -- I44_Y fo2 L3(2,1,0)
    N506 := (DataBReg(25) and DataAReg(25))
         or (DataBReg(25) and Gene(24))
         or (DataAReg(25) and Gene(24));
  -- I45_Y fo3 L3(2,2)
    N507 := prop(25) and Prop(24);

-- Sum27
  -- I270_Y_0_a2_0_1 fo3 L3(2,2)
    N87_1 := Prop(25) and Prop(26);
  -- I42_Y_0_o2 fo2 L3(2,2,2)
    N504 := (Gene(25) and Prop(26)) or Gene(26);

-- Sum28
  -- I40_Y fo2 L3(2,1,0)
    N502 := (DataBReg(27) and DataAReg(27))
         or (DataBReg(27) and Gene(26))
         or (DataAReg(27) and Gene(26));
  -- I41_Y fo3 L3(2,2)
    N503 := Prop(27) and Prop(26);

-- Sum29
  -- I39_Y_i_o2 fo4 L3(2,2)
    N78 := Prop(27) and Prop(28);
  -- I38_Y_0_o2 fo2 L3(2,2,2)
    N500 := (Gene(27) and Prop(28)) or Gene(28);

-- Sum30
  -- I36_Y fo2 L3(2,1,0)
    N498 := (DataBReg(29) and DataAReg(29))
         or (DataBReg(29) and Gene(28))
         or (DataAReg(29) and Gene(28));
  -- I37_Y fo4  L3(2,2)
    N499 := Prop(29) and Prop(28);

-- Sum31
  -- I35_Y fo2 L3(2,2)
    N497 := Prop(30) and Prop(29);
  -- I268_Y_0 fo1 L3(2,2,2)
    I268Y0 := (Gene(29) and Prop(30)) or Gene(30);

-- Sum32
  -- I33_Y fo3 L3(2,2)
    N495 := Prop(31) and Prop(30);
  -- I300_Y_0 fo1   L3(2,1,0)
    I300Y0 := (DataBReg(31) and DataAReg(31))
           or (DataBReg(31) and Gene(30))
           or (DataAReg(31) and Gene(30));
  -- I300_Y_1 fo1 L4(3,3)
    I300Y1 := (N495 and N498) or I300Y0;
end if;

------------------------------------------------------------
------------------------------------------------------------
-- L4

-- Sum3
  -- I210_Y_0_o3_0 fo4 L4(3,2,2)
    N619 := (N621_dup and Prop(2)) or Gene(2);
  -- I210_Y_0_o3_0 fo4 L4(3,2,2)
    N619_dup := N621_dup and Prop(2);

-- Sum4
  -- I152_Y fo2 L4(3,3,3)
    N617 := (N551 and N554) or N550;
  -- I153Y fo2 L4(3,3)
    N618 := N555 and N551;

-- Sum6
  -- I148_Y fo2 L4(3,3,3)
    N613 := (N547 and N550) or N546;
  -- I149_Y fo2 L4(3,3)
    N614 := N547 and N551;

-- Sum7
  -- I146_Y fo4 L4(3,3,3)
    N611 := (N548i and N_42) or I146Y0;
  -- I147_Y fo2 L4(2,2,3)
    N612 := Prop(4) and Prop(3) and N_42;

-- Sum8
  -- I144_Y fo2 L4(3,3,3)
    N609 := (N543 and N546) or N542;
  -- I145_Y fo3 L4(3,3)
    N610 := N547 and N543;

-- Sum10
  -- I140_Y fo2 L4(3,3,3)
    N605 := (N539 and N542) or N538;
  -- I141_Y fo3 L4(3,3)
    N606 := N543 and N539;

-- Sum11
  -- I138_Y fo2 L4(3,3,3)
    N603 := (N540i and N_40) or I138Y0;
  -- I139_Y fo3 L4(3,3)
    N604 := I139Y0 and N_40;

-- Sum12
  -- I136_Y fo2 L4(3,3,3
    N601 := (N535 and N538) or N534;
  -- I137_Y fo4 L4(3,3)
    N602 := N539 and N535;

-- Sum14
  -- I132_Y fo2 L4(3,3,3)
    N597 := (N531 and N534) or N530;
  -- I133_Y fo3 L4(3,3)
    N598 := N535 and N531;

-- Sum15
  -- I130_Y fo2 L4(3,3,3)
    N595 := (N532 and N38i0) or I130Y0;
  -- I131_Y fo3 L4(2,2,3)
    N596 := Prop(11) and Prop(12) and N38i0;
    
-- Sum16
  -- I128_Y fo2 L4(3,3,3)
    N593 := (N527 and N530) or N526;
  -- I129_Y fo5 L4(3,3)
    N594 := N531 and N527;
    N594_dup := N531 and N527;

if YASEP_SIZE=32 then

-- Sum18
  -- I124_Y fo2 L4(3,3,3)
    N589 := (N523 and N526) or N522;
  -- I125_Y fo3 L4(3,3)
    N590 := N527 and N523;

-- Sum19
  -- I122_Y fo2 L4(3,3,3)
    N587 := (N524 and N_36) or I122Y0;
  -- I123_Y fo3 L4(3,3)
    N588 := N105_1 and N_36;

-- Sum20
  -- I120_Y fo2 L4(3,3,3)
    N585 := (N519 and N522) or N518;
  -- I121_Y fo5 L4(3,3)
    N586 := N523 and N519;

-- Sum22
  -- I116_Y fo2 L4(3,3,3)
    N581 := (N515 and N518) or N514;
  -- I117_Y fo3 L4(3,3)
    N582 := N519 and N515;

-- Sum23
  -- I114_Y fo2 L4(3,3,3)
    N579 := (N516 and N89_1) or N512;
  -- I115_Y fo4 L4(3,3)
    N580 := (N109_1 and N89_1);

-- Sum24
  -- I112_Y fo2 L4(3,3,3)
    N577 := (N511 and N514) or N510;
  -- I113_Y fo3 L4(3,3)
    N578 := N515 and N511;

-- Sum25
  -- I274_Y_i_a3_0_0 fo3 L4(3,3)
    I274Yia300 := N89_1 and N79;
  -- I274_Y_i_o3_0 fo3 L4(3,3,3)
    I274Yio30 := not( (N512 and N79) or N508 );
  -- I358_Y_m3_e_0 fo L4(3,3)
    I358Ym3e0 := not(N516) and N109_1;
  -- d_m1_e fo1 L4(3,3)
    dm1e := N89_1 and N79 and N516;

-- Sum26
  -- I108_Y fo2 L4(3,3,3)
    N573 := (N507 and N510) or N506;
  -- I109_Y fo3 L4(3,3)
    N574 := N511 and N507;

-- Sum27
  -- I106_Y fo2  L4(3,3,3)
    N571 := (N87_1 and N508) or N504;
  -- I107_Y fo3  L4(3,3)
    N572 := N87_1 and N79;

-- Sum28
  -- I104_Y fo2 L4(3,3,3)
    N569 := (N503 and N506) or N502;
  -- I105_Y fo3 L4(3,3)
    N570 := N507 and N503;

-- Sum29
  -- I270_Y_0_a2_0_0 fo2 L4(3,3)
    I270Y0a200 := N87_1 and N78;
  -- I270_Y_0 fo1 L4(3,3,3)
    I270Y0 := (N504 and N78) or N500;
  -- I270_Y_0_a1_0 fo1 L4(3,3,3)
    I270Y0a10 := N87_1 and N78 and N516;

-- Sum30
  -- I101_Y fo2 L4(3,3)
    N566 := N503 and N499;
  -- I269_Y_0 L4(3,3,3)
    I269Y0 := (N502 and N499) or N498;

-- Sum31
  -- I99_Y fo2 L4(3,3)
    N564 := N497 and N78;
  -- I268_Y_1 fo1 L4(3,3,3)
    I268Y1 := (N497 and N500) or I268Y0;

-- Sum32
  -- I97_Y fo1  L5(4,3,3)
    N562 := N569 and N495 and N499;
  -- I159_Y fo3  L5(4,3,3)
    N630   := N570 and N495 and N499;

end if;

------------------------------------------------------------
------------------------------------------------------------
-- L5

-- Sum2
  -- I335_Y L5(2,4)
    Sum2rr := Comp(2) xor N621;

-- Sum4
  -- I211_Y fo4 L5(1,4,4)
    N685 := (DataBReg(0) and N618) or N617;

-- Sum6
  -- I209_Y fo3 L5(4,4,4)
    N681 := (N614 and N621) or N613;

-- Sum7
  -- I208_un1_Y fo3 L5(4,4)
    I208un1Y := (N619_dup or Gene(2)) and N612;

-- Sum8
  -- I206 fo2 L5(4,4,4)
    N677 := (N610 and N617) or N609;
  -- I207_Y fo2 L5(4,4)
    N678 := N618 and N610;

-- Sum10
  -- I202_Y fo2 L5(4,4,4)
    N673 := (N606 and N613) or N605;
  -- I203_Y fo2 L5(4,4)
    N674 := N614 and N606;

-- Sum11
  -- I200_Y fo2 L5(4,4,4)
    N671 := (N604 and N611) or N603;
  -- I201_Y fo2 L5(4,4)
    N672 := N612 and N604;-- Sum12

-- Sum13
  -- I198Y fo2 L5(4,4,4)
    N669 := (N602 and N609) or N601;
  -- I199_Y fo2 L5(4,4)
    N670 := N610 and N602;
    
-- Sum14
  -- I194_Y fo2 L5(4,4,4)
    N665 := (N598 and N605) or N597;
  -- I195_Y fo2 L5(4,4)
    N666 := N606 and N598;

-- Sum15
  -- I192_Y fo4 L5(4,4,4)
    N663 := (N596 and N603) or N595;
  -- I193_Y fo2 L5(4,4)
    N664 := N604 and N596;

-- Sum16
  -- I190_Y fo2 L5(4,4,4)
    N661 := (N594 and N601) or N593;
  -- I191_Y fo3 L5(4,4)
    N662 := N602 and N594_dup;
    N662_dup := (N539 and N535) and N594_dup;

if YASEP_SIZE=32 then

-- Sum18
  -- I186_Y fo2 L5(4,4,4)
    N657 := (N590 and N597) or N589;
  -- I187_Y fo3 L5(4,4)
    N658 := N598 and N590;

-- Sum19
  -- I185_Y fo3 L5(4,4)
    N656 := N596 and N588;
  -- I184_Y fo2 L5(4,4,4)
    N655 := (N588 and N595) or N587;

-- Sum20
  -- I182_Y fo2 L5(4,4,4)
    N653 := (N586 and N593) or N585;
  -- I279_un1_Y_0 fo1 L5(4,4,4)
    I279un1Y0 := N586 and N594 and N670;

-- Sum22
  -- I178_Y fo2 L5(4,4,4)
    N649 := (N582 and N589) or N581;
  -- I179_Y fo3 L5(4,4)
    N650 := N590 and N582;

-- Sum23
  -- I176_Y fo2 L5(4,4,4)
    N647 := (N580 and N587) or N579;
  -- I177_Y fo3 L5(4,4)
    N648 := N588 and N580;

-- Sum24
  -- I174_Y fo2 L5(4,4,4)
    N645 := (N578 and N585) or N577;
  -- I175_Y fo4 L5(4,4)
    N646 := N586 and N578;

-- Sum25
  -- I358_Y_m3_e fo1 L5(4,4,4)
    I358YN4i := I274Yia300 and I358Ym3e0 and I274Yio30;
  -- d_b0_0_x2 fo1  L5(4,4,2)
    N4i := (not(dm1e) and I274Yio30) xor Comp(25);

-- Sum26
  -- I171_Y fo2 L5(4,4)
    N642 := N582 and N574;
  -- I273_Y_0 fo1 L5(4,4,4)
    I273Y0 := (N574 and N581) or N573;

-- Sum27
  -- I272_Y_0 fo1 L5(4,4,4)
    I272Y0 := (N572 and N579) or N571;
  -- I169_Y fo2 L5(4,4)
    N640 := N580 and N572;

-- Sum28
  -- I167_Y fo2 L5(4,4)
    N638 := N578 and N570;
  -- I271_Y_0 fo1 L5(4,4,4)
    I271Y0 := (N570 and N577) or N569;
  -- I271_un1_Y_0 fo1 L5(4,4)
    I271un1Y0 := N586 and N594;

-- Sum29
  -- I270_Y_0_a0_1 fo1 L5(4,3,4)
    I270Y0a01 := I270Y0a200 and N109_1 and I274Yia300;
  -- I270_Y_1 qt7+1 fo1  L5(4,4,4)
    I270Y1 := (I274Yia300 and I270Y0a10) or I270Y0;

-- Sum30
  -- I163_Y fo2 L5(4,4)
    N634 := N574 and N566;
  -- I269_Y_1  L5(4,4,4)
    I269Y1 := (N573 and N566) or I269Y0;

-- Sum31
  -- I161_Y fo2 L5(4,4)
    N632 := N572 and N564;
  -- I268_Y_2 fo1 L5(4,4,4)
    I268Y2 := (N571 and N564) or I268Y1;

end if;

------------------------------------------------------------
------------------------------------------------------------
-- L6

-- Sum3
  -- I336_Y L6(2,5)
    Sum3r := Comp(3) xor N619;

-- Sum4
  -- I337_Y L6(2,5)
    Sum4r := Comp(4) xor N685;

-- Sum5
  -- I210_Y_0_a3 fo1 L6(5,2,2)
    I210Y0a3 := (N685 and Prop(4)) or Gene(4);

-- Sum6
  -- I339_Y L6(2,5)
    Sum6r := Comp(6) xor N681;

-- Sum8
  -- I257_Y fo3 L6(5,5)
    N926 := (DataBReg(0) and N678) or N677;

-- Sum10
  -- I255_Y fo2 L6(5,4,5)
    N920 := (N674 and N621) or N673;

-- Sum11
  -- I254_Y fo2 L6(5,5,5)
    N917 := (N619 and N672) or N671;

-- Sum12
  -- I253_Y fo3 L6(5,5,5)
    N914 := (N670 and N685) or N669;
    
-- Sum14
  -- I251_Y fo2 L6(5,5,5)
    N908 := (N666 and N681) or N665;

-- Sum15
  -- I250_un1_Y fo3 L6(5,4,5)
    I250un1Yi := (I208un1Y or N611) and N664;

-- Sum16
  -- I249_Y fo2 L6(5,5,5)
    N731 := (N662 and N677) or N661;
  -- I249_Y fo2 L6(5,5,5)
    N732 := N678 and N662 and DataBReg(0);


if YASEP_SIZE=32 then

-- Sum18
  -- I244_un1_Y fo1 L6(4,5)
    I244un1Y := N673 and N658;
  -- I281_un1_Y_0 fo1 L6(4,5)
    I281un1Y0 := N621 and N674;

-- Sum19
  -- I280_un1_Y fo4 L6(5,5,5)
    I280un1Y := N656 and N672 and N619;
  -- I280_Y_0 fo4 L6(5,5,5)
    I280Y0 := (N656 and N671) or N655;

-- Sum20
  -- I279_Y_0 fo1 L6(5,5,5)
    I279Y0 := (N685 and I279un1Y0) or N653;
  -- I240_un1_Y fo1 L6(5,4,5)
    I240un1Y := N586 and N594 and N669;

-- Sum22
  -- I277_un1_Y fo1 L6(5,5,5)
    I277un1Y := N650 and N666 and N681;
  -- I277_Y_0 fo1 L6(5,5,5)
    I277Y0 := (N650 and N665) or N649;

-- Sum23
  -- I276_un1_Y_0 fo1 L6(5,5)
    I276un1Y0 := N664 and N648;
  -- I276_Y_0 fo1 L6(5,5,5)
    I276Y0 := (N648 and N663) or N647;

-- Sum24
  -- I275_Y_0 fo1 L6(5,5,5)
    I275Y0 := (N646 and N661) or N645;

-- Sum26
  -- I273_Y_1 fo1 L6(5,5,5)
    I273Y1 := (N642 and N657) or I273Y0;

-- Sum27
  -- I272_un1_Y_0 fo1 L6(5,5)
    I272un1Y0 := N656 and N640;
  -- I272_Y_1 fo1 L6(5,5,5)
    I272Y1 := (N640 and N655) or I272Y0;

-- Sum28
  -- I271_Y_1 fo1 L6(5,5,5)
    I271Y1 := (N638 and N653) or I271Y0;

-- Sum29
  -- I270_Y_2 fo1 L6(4,4,5)
    I270Y2 := (I270Y0a200 and not(I274Yio30)) or I270Y1;

-- Sum30
  -- I269_Y_2 L6(5,5,5)
    I269Y2 := (N634 and N649) or I269Y1;

-- Sum31
  -- I268_un1_Y_0 fo1 L6(5,5)
    I268un1Y0 := N632 and N648;

-- Sum32
  -- I300_un1_Y_0 fo1 L6(1,5, 5)
    I300un1Y0 := DataBReg(0) and N630 and N646;
  -- I300_Y_3 fo1 L6(5,5)
    I300Y3 := N645 and N630;
end if;

------------------------------------------------------------
------------------------------------------------------------
-- ROP2

 -- L3 : Mux1
  if (ROP2_S1='1') then
    ROP2_RS1 := Gene; -- and
  else
    ROP2_RS1 := Comp; -- xor
  end if;

 -- L4 : Mux2
  if (ROP2_S2='1') then
    ROP2_RS2 := Prop; -- or
  else
    ROP2_RS2 := ROP2_RS1;
  end if;

  -- L4 : OR-reduction :
  if ROP2_RS1=(ROP2_RS1'range=>'0') then
    ROP2_c := '0';
  else
    ROP2_c := '1';
  end if;

  -- L5 : XOR2 (invert the result if ROP2_X2 is set)
  ROP2_RX2 := ROP2_RS2 xor (ROP2_RX2'range => ROP2_X2);

------------------------------------------------------------
------------------------------------------------------------
-- pipeline gate #2 :
  if (PIPELINED=0) or (clk'event and clk='1') then --  and add_EN2='1'
    Sum(0)   <= Sum0rr;
    Sum(1)   <= Sum1rr;
    Sum(2)   <= Sum2rr;
  --L2
    Comp5r   := Comp(5);
    Comp7r   := Comp(7);
    Comp8r   := Comp(8);
    Prop8r   := Prop(8);
    Gene8r   := Gene(8);
    Comp9r   := Comp(9);
    Comp10r  := Comp(10);
    Comp11r  := Comp(11);
    Comp12r  := Comp(12);
    Gene12r  := Gene(12);
    Prop12r  := Prop(12);
    Comp13r  := Comp(13);
    Comp14r  := Comp(14);
    Comp15r  := Comp(15);
    if YASEP_SIZE=32 then
      Comp16r  := Comp(16);
      Comp17r  := Comp(17);
      Comp18r  := Comp(18);
      Comp19r  := Comp(19);
      Comp20r  := Comp(20);
      Comp21r  := Comp(21);
      Comp22r  := Comp(22);
      Comp23r  := Comp(23);
      Comp24r  := Comp(24);
      Comp25r  := Comp(25);
      Comp26r  := Comp(26);
      Comp28r  := Comp(28);
      Comp29r  := Comp(29);
      Comp30r  := Comp(30);
      Comp31r  := Comp(31);

    --L3
      N109_1r  := N109_1;
      N105_1r  := N105_1;
      N524r    := N524;
      N516r    := N516;
    else
      Comp16r  := '0'; -- necessary for the carry in 16-bit mode
    end if;

    --L4
    N611r    := N611;
    if YASEP_SIZE=32 then
      I300Y1r  := I300Y1;

    --L5
      N657r      := N657;
    end if;
    N663r      := N663;
    I208un1Yr  := I208un1Y;
    if YASEP_SIZE=32 then
      I271un1Y0r := I271un1Y0;
      I270Y0a01r := I270Y0a01;
      I358YN4ir  := I358YN4i;
      I268Y2r    := I268Y2;
      N4ir       := N4i;
      N562r      := N562;
      N630r      := N630;
      N632r      := N632;
      N634r      := N634;
      N638r      := N638;
      N642r      := N642;
      N646r      := N646;
      N647r      := N647;
      N650r      := N650;
      N658r      := N658;
      N662r      := N662;
      N678r      := N678;
      N662_dupr  := N662_dup;
    end if;
    -- L6
    Sum(3)     <= Sum3r;
    Sum(4)     <= Sum4r;
    I210Y0a3r  := I210Y0a3;
    Sum(6)     <= Sum6r;
    N926r      := N926;
    N920r      := N920;
    N917r      := N917;
    N914r      := N914;
    N908r      := N908;
    N731r      := N731;
    N732r      := N732;
    I250un1Yir := I250un1Yi;
    if YASEP_SIZE=32 then
      I244un1Yr  := I244un1Y;
      I281un1Y0r := I281un1Y0;
      I280un1Yr  := I280un1Y;
      I280Y0r    := I280Y0;
      I279Y0r    := I279Y0;
      I240un1Yr  := I240un1Y;
      I277un1Yr  := I277un1Y;
      I277Y0r    := I277Y0;
      I276un1Y0r := I276un1Y0;
      I276Y0r    := I276Y0;
      I275Y0r    := I275Y0;
      I273Y1r    := I273Y1;
      I272un1Y0r := I272un1Y0;
      I272Y1r    := I272Y1;
      I271Y1r    := I271Y1;
      I270Y2r    := I270Y2;
      I269Y2r    := I269Y2;
      I268un1Y0r := I268un1Y0;
      I300un1Y0r := I300un1Y0;
      I300Y3r    := I300Y3;
    end if;

    ROP2_OUT <= ROP2_RX2;
    ROP2_combine <= ROP2_c;
	end if;

--  if (clk'event and clk='1') then
--    add_EN2 := add_en;
--  end if;


--------------------------------------------------------------
------------------------------------------------------------
-- L7

-- Sum5
  -- I338_Y L7(2,6)
    Sum(5) <= Comp5r xor I210Y0a3r;

-- Sum7
  -- I340_Y L7(6,4,2)
    Sum(7) <= (I208un1Yr or N611r) xor Comp7r;

-- Sum8
  -- I341_Y L7(2,6)
    Sum(8) <= Comp8r xor N926r;

-- Sum9
  -- I256_Y_0_a3 fo1 L7(6,2,2)
    I256Y0a3 := (N926r and Prop8r) or Gene8r;

-- Sum10
  -- I343_Y L7(2,6)
    Sum(10) <= Comp10r xor N920r;

-- Sum11
  -- I344_Y L7(2,6)
    Sum(11) <= Comp11r xor N917r;

-- Sum12
  -- I345_Y L7(2,6)
    Sum(12) <= Comp12r xor N914r;

-- Sum13
  -- I252_Y_0_o2 fo1 L7(6,2,2)
    N911i := (Prop12r and N914r) or Gene12r;

-- Sum14
  -- I347_Y L7(6,2)
    Sum(14) <= Comp14r xor N908r;

-- Sum16
  -- I349_Y L7(6,6,2)
    Sum(16) <= (N731r or N732r) xor Comp16r;

if YASEP_SIZE=32 then

-- Sum17
  -- 128_Y_0_a3 fo1 L7(6,5,3)
    I282Y0a := (I250un1Yir or N663r) and N105_1r;

-- Sum18
  -- I281_Y_0 fo1 L7(5,6,5)
    I281Y0 := (N658r and I281un1Y0r) or N657r;

-- Sum19
  -- I352_Y L7(6,6,2)
    Sum(19+dummy_base) <= (I280un1Yr or I280Y0r) xor Comp19r;

-- Sum20
  -- I353_Y L7(6,6,2)
    Sum(20+dummy_base) <= (I240un1Yr or I279Y0r) xor Comp20r;

-- Sum21
  -- I278_Y_0_a3 fo1 L7(6,6,3)
    I278Y0a3 := (I280un1Yr or I280Y0r) and N109_1r;

-- Sum22
  -- I355_Y L7(6,6,2)
    Sum(22+dummy_base) <= (I277un1Yr or I277Y0r) xor Comp22r;

-- Sum23
  -- I276_un1_Y fo1 L7(6,4,6)
    I276un1Y := (I208un1Yr or N611r) and I276un1Y0r;

-- Sum24
  -- I275_un1_Y fo1 L7(5,6,5)
    I275un1Y := N662r and N926r and N646r;

-- Sum25
  -- r_m1 fo1 L7(6,6,2)
    rm1 := (I280un1Yr or I280Y0r) xor Comp25r;

-- Sum26
  -- I273_un1_Y fo1 L7(5,5,6)
    I273un1Y := N658r and N642r and N920r;

-- Sum27
  -- I272_Y fo1 L7(6,6,6)
    N869i := (N917r and I272un1Y0r) or I272Y1r;

-- Sum28
  -- I271_Y  L7(6,5,5)
    N866i := (N914r and I271un1Y0r and N638r) ;

-- Sum29
  -- I270_Y_0_a0 fo1  L7(6,6,5)
    I270Y0a0 := (I280un1Yr or I280Y0r) and I270Y0a01r;

-- Sum30
  -- I269_un1_Y    L7(5,5,6)
    I269un1Y := N650r and N634r and N908r;

-- Sum31
  -- I268_un1_Y fo1 L7(5,7,6)
    I268un1Y := (N663r or I250un1Yir) and I268un1Y0r;
  -- I268_Y_3 fo1 L7(5,5,6)
    I268Y3 := (N632r and N647r) or I268Y2r;

-- Sum32
  -- L7(5,4,6)
    Sum32_or := N562r or I300Y1r or I300Y3r;
  -- I266_un1_Y fo1  L7(6,5,5)
    I266un1Y := N731r and N630r and N646r;
  -- I300_Y_4 fo1  L7(5,5,6)
    I300Y4 := N678r and N662_dupr and I300un1Y0r;

end if;

------------------------------------------------------------
------------------------------------------------------------
-- L8

-- Sum9
  -- I342_Y  L8(7,2)
    Sum(9) <= I256Y0a3 xor Comp9r;

-- Sum13
  -- I346_Y L8(2,7)
    Sum(13) <= Comp13r xor N911i;

-- Sum15
  -- I348_Y L8(5,7,2)
    Sum(15) <= (N663r or I250un1Yir) xor Comp15r;

if YASEP_SIZE=32 then

-- Sum17
  -- I350_Y L8(7,3,2)
    Sum(17+dummy_base) <= (I282Y0a or N524r) xor Comp17r;

-- Sum18
  -- I351_Y L8(6,7,2)
    Sum(18+dummy_base) <= (I244un1Yr or I281Y0) xor Comp18r;

-- Sum21
  -- I354_Y L8(7,3,2)
    Sum(21+dummy_base) <= (I278Y0a3 or N516r) xor Comp21r;

-- Sum22
  -- I356_Y L8(7,6,2)
    Sum(23+dummy_base) <= (I276un1Y or I276Y0r) xor Comp23r;

-- Sum24
  -- I357_Y L8(7,6,2)
    Sum(24+dummy_base) <= (I275un1Y or I275Y0r) xor Comp24r;

-- Sum25
  -- I358_Y_1 L8(5,5,7)
    if (I358YN4ir = '0') then
      Sum(25+dummy_base) <= not(N4ir);
    else
      Sum(25+dummy_base) <= rm1;
    end if;

-- Sum26
  -- I359_Y L8(7,6,2)
    Sum(26+dummy_base) <= (I273un1Y or I273Y1r) xor Comp26r;

-- Sum27
  -- I360_Y_0 fo1 L8(7,7,7)
    Sum(27+dummy_base) <= Comp(27) xor N869i;

-- Sum28
  -- I361_Y L8(2,7,6)
    Sum(28+dummy_base) <= Comp28r xor (N866i or I271Y1r);

-- Sum29
  -- I362_Y L8(7,6,2)
    Sum(29+dummy_base) <= (I270Y0a0 or I270Y2r) xor Comp29r;

-- Sum30
  -- I363_Y L8(7,6,2)
    Sum(30+dummy_base) <= (I269un1Y or I269Y2r) xor Comp30r;

-- Sum31
  -- I364_Y L8(7,7,2)
    Sum(31+dummy_base) <= (I268un1Y or I268Y3) xor Comp31r;

-- Sum32
  -- I365_Y L8(7,7,7)
    Sum(32+dummy_base) <= I266un1Y or I300Y4 or Sum32_or;
end if;

------------------------------------------------------------

end process;
end ActelProASIC3;
