-- SPI_unit.vhd
-- created 20080823 by Yann Guidon
-- This is a SPI protocol master unit for YASEP in an Actel FPGA.
-- This design is mostly asynchronous, or at least decoupled,
-- and can use external an clock source (think ENC28J60)

-- Fixme :  ClkPhase not working yet

-- Description :
--  There are 2 registers mapped to the SR space of YASEP.
--  They are accessed by the GET and PUT instructions, that latch the address
--  and the written data on DataIn and Addr0.
--  A write strobe (WR) should also be externally provided.
--
-- * Address 0 is the control register, Read/write (except RDY: read only)
--    bit 0 : RDY (Ready): when=1, then a new transaction can be started.
--       when 0, a transaction is taking place.
--    bit 1 : SPI_EN : when 0, all functions are disabled (default after reset)
--       Set to 1 before starting a new transaction.
--    bit 2: Len8/16 : 0=>8bits (default), 1=>16bits per word/transaction 
--    bits 3 & 4 : Chip Select 1 to 4 (0 at reset)
--       This selects 1 out of 4 ENable outputs.
--       The selected output is asserted when SPI_EN is 1
--    bit 8 : CLKphase (0 at reset)
--    bit 9 : CLKpolarity (0 at reset)
--    bit 10 : EXTernal clock (0 at reset)
--    bits 11 to 15 : Clock Division (00000 at reset : slowest clock)
--       11111 => clk/2
--       11110 => clk/4
--       11101 => clk/6
--       ....  clk/(2x(-clkdiv))
--       00001 => clk/62
--       00000 => clk/64  ==> so when CPUclk=64MHz, SCK=1MHz (for example)
-- * Address 1 is the data register.
--  - when read, it provides the contents of the input shift register.
--  - when written, the data are transfered to the output register and a new transaction starts.
--
-- How to use it :
--  - configure the control/status register :
--     - set SPI_EN to 1, and Len8/16, CS, CLKpol, CLKph, CLKdiv to the correct values
--         MOV 0002h,r0 ; EN=1, clkdiv=pol=phase=0, CS1 enabled, len=8 bits
--         PUT r0,SPI_CTL
--
--  - Write to the data register to start a transaction.
--      when read only is desired, the value of the data register can be set to 0 or disregarded.
--         MOV 1234h, r0
--         PUT r0,SPI_DATA
--
--  - poll the value of RDY in the control register
-- label_poll:  GET SPI_CTL,R0
--              JE label_poll
--
--  - If data is expected from the slave device, read it on SPI_DATA
--         GET SPI_DATA, R0
--
--  - To end the transaction, clear the SPI_EN bit (so CSx will go high again).
--      It more data have to be read or written, then read/write the
--      data register again.
--
--  - when the SPI interface becomes useless, don't forget to turn it off :
--         xor R0,R0
--         PUT r0,SPI_CTL
--


library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity SPI_unit is
  port(
    -- CPU side
    clk, reset, Addr0, Wr : in std_logic;
    DataIn  :  in std_logic_vector(15 downto 0);
    DataOut : out std_logic_vector(15 downto 0);
    
    -- SPI side :
    CS1, CS2, CS3, CS4, Sck, SDout : out std_logic;
    SDin, ext_clk : in std_logic);
end SPI_unit;

architecture ActelProAsic3 of SPI_unit is
  signal ShRegIn, ShRegOut : std_logic_vector(15 downto 0);
  signal CS : std_logic_vector(1 downto 0);
  signal cnt5, clkdiv, divcounter : std_logic_vector(4 downto 0);
  signal ClkPol, ClkPhase, Len8_16, RDY, preRDY, SPI_en,  -- configuration bits
      lCK, lCK_en, lCKok, eCK_en, CK_data, CK_ctrl, EXTck, gl1, gl2 : std_logic;

begin

------------------------------ CPU interface ------------------------------

  CS1 <= not(CS(0)) and not(CS(1)) and SPI_EN;
  CS2 <=     CS(0)  and not(CS(1)) and SPI_EN;
  CS3 <= not(CS(0)) and     CS(1)  and SPI_EN;
  CS4 <=     CS(0)  and     CS(1)  and SPI_EN;

  -- CPU read :
  DataOut <=  ShRegIn  when (Addr0='1') else
      clkdiv & EXTck & ClkPol & ClkPhase      -- high byte :  clock management
     & "000" & CS & Len8_16 & SPI_en & RDY ;  -- lower byte : transfer & frame management

-- clock gating of the registers
-- (because both ENABLE and RESET are not possible at the same time) :
-- WR must be valid only during one CLK cycle, the cycle starts low
-- and Addr0 should be active during at least 1 cycle
  CK_data <= clk and     Addr0  and Wr;
  CK_ctrl <= clk and not(Addr0) and Wr; 

  Data_Write: process(CK_data) is
  begin
    if (CK_data'event and CK_data='1') then    -- front descendant ?
      ShRegOut <= DataIn;
    end if;
  end process;

  Control_Write: process(reset, CK_ctrl) is
  begin
    if (reset='1') then
      ClkPol   <= '0';   -- SPI mode 0
      ClkPhase <= '0';
      CS       <= "00"; -- CS1 by default
      SPI_en   <= '0';  -- prevent the shift clock from starting
      Len8_16  <= '0';  -- default width is 8 bits
      EXTck    <= '0';  -- internal clock by default
      clkdiv   <= (others=>'0'); -- slowest clock
    else
      if (CK_ctrl'event and CK_ctrl='1') then  -- front descendant ?
        -- high byte : clock management
        clkdiv   <= DataIn(15 downto 11);
        EXTck    <= DataIn(10);
        ClkPol   <= DataIn(9);
        ClkPhase <= DataIn(8);
        -- lower byte : transfer & frame management :
        CS       <= DataIn(4 downto 3);
        Len8_16  <= DataIn(2);
        SPI_en   <= DataIn(1);
      end if;
    end if;
  end process;


------------------------------ Clock conditionning ------------------------------

  -- clock divider
  clock_process : process(SPI_en, clk)
    variable t : std_logic_vector(5 downto 0);
  begin
    if (clk'event and clk='1' and SPI_en='1') then
      t := std_logic_vector( unsigned('0' & divcounter) + 1 );  -- increment the counter
      if t(t'left)='1' then -- if (expected) overflow then toggle lCK
        divcounter <= clkdiv;
        lCK <= not(lCK);
      else
        divcounter <= t(divcounter'range);  -- Just update the counter
      end if;
    end if;
  end process;


-- Clock resynchronization :
-- see http://www.design-reuse.com/articles/5827/techniques-to-make-clock-switching-glitchfree.html
-- I added some "enables"
  Sync: process(reset, CK_data, lCK, ext_clk) is
  begin
    if (reset='1' or CK_data='1') then
      gl1 <= '0';
      lCK_en <= '0';
      gl2<='0';
      eCK_en<='0';
    else
      if (lCK'event and lCK='1') then
        gl1 <= SPI_en and not(EXTck) and not(eCK_en) and not(preRDY);
      end if;
      if (lCK'event and lCK='0') then
        lCK_en <= gl1;
      end if;
      if (ext_clk'event and ext_clk='1') then
        gl2 <= SPI_en and EXTck and not(lCK_en) and not(preRDY);
      end if;
      if (ext_clk'event and ext_clk='0') then
        eCK_en <= gl2;
      end if;
    end if;
  end process;
  
  lCKok <= (lCK and lCK_en) or (ext_clk and eCK_en);

  Sck <= lCKok xor ClkPol;


------------------------------ Data transmission ------------------------------

  -- The receive part is ... simple.
  -- this should synthesize to a pretty minimal
  -- circuit made of 16xFF and one AND gate
  Receive: process(lCKok) is
  begin
    if (lCKok'event and lCKok='0') then
      -- masks the MSB of the shift register in 8-bit mode
      -- with 1 gate instead of 8
      ShRegIn(15 downto 9) <= ShRegIn(14 downto 8);
      ShRegIn(          8) <= ShRegIn(          7) and Len8_16;
      ShRegIn( 7 downto 1) <= ShRegIn( 6 downto 0);
      ShRegIn(          0) <= SDin;
    end if;
  end process;

  -- The emitter is mostly a "serialiser".
  -- A shift register approach is too .. huh...
  -- troublesome because of the different clocks.
  -- So the data register is latched normally
  -- and the serial output is just selected by a
  -- Gray counter (to avoid output glitches).
  Gray: process(lCKok, Addr0, Wr, Len8_16) is
    variable nC2, nC3 : std_logic;
    variable t_cnt5: std_logic_vector(4 downto 0);
  begin
    if (Addr0='1' and Wr='1') then
      cnt5 <= (others=>'0');
      nC2 :='1';
      nC3 :='1';    
    else
      if (lCKok'event and lCKok='1') then
        t_cnt5 := cnt5;
        if (Len8_16='0') then
          t_cnt5(2) := nC2;
          t_cnt5(3) := nC3;
        end if;
        case t_cnt5 is
          -- ok i know... it's bloated but Synplicity makes an excellent job,
          -- reducing this to <10 simple logic gates
          when "00000" => t_cnt5 := "00001"; -- 0  (init state when Len8_16=1 - 16-bit mode)
          when "00001" => t_cnt5 := "00011"; -- 1
          when "00011" => t_cnt5 := "00010"; -- 2
          when "00010" => t_cnt5 := "00110"; -- 3
          when "00110" => t_cnt5 := "00111"; -- 4
          when "00111" => t_cnt5 := "00101"; -- 5
          when "00101" => t_cnt5 := "00100"; -- 6
          when "00100" => t_cnt5 := "01100"; -- 7
          when "01100" => t_cnt5 := "01101"; -- 8  (init state when Len8_16=0 : 8-bit mode)
          when "01101" => t_cnt5 := "01111"; -- 9
          when "01111" => t_cnt5 := "01110"; -- 10
          when "01110" => t_cnt5 := "01010"; -- 11
          when "01010" => t_cnt5 := "01011"; -- 12
          when "01011" => t_cnt5 := "01001"; -- 13
          when "01001" => t_cnt5 := "01000"; -- 14
          when "01000" => t_cnt5 := "11000"; -- 15
          when others  => t_cnt5 := "11000"; -- don't continue if overflow/RDY
              -- cool side effect : state #16 maintains the last bit on Q
        end case;
        cnt5 <= t_cnt5;
        nC2  := t_cnt5(2);
        nC3  := t_cnt5(3);
      end if;
    end if;
  end process;

  RDY <= cnt5(4);
  preRDY <= '1' when (cnt5(4)='1' or cnt5="01000") else '0';
   -- preRDY provides an advance signal for the anti-glitch circuit
   -- so the clock will be held low when the counter expires.


  -- The serializer is just a binary tree of MUXes.
  -- Except that the entries are quite... shuffled.
  serialize: process(cnt5, ShRegOut) is
    variable t8 : std_logic_vector(7 downto 0);
    variable t4 : std_logic_vector(3 downto 0);
    variable t2 : std_logic_vector(1 downto 0);
  begin

    if (cnt5(0)='0') then    -- Gray:
      t8(0) := ShRegOut(15); -- 0000
      t8(1) := ShRegOut(0);  -- 1000
      t8(2) := ShRegOut(8);  -- 0100
      t8(3) := ShRegOut(7);  -- 1100
      t8(4) := ShRegOut(12); -- 0010
      t8(5) := ShRegOut(3);  -- 1010
      t8(6) := ShRegOut(11); -- 0110
      t8(7) := ShRegOut(4);  -- 1110
    else
      t8(0) := ShRegOut(14); -- 0001
      t8(1) := ShRegOut(1);  -- 1001
      t8(2) := ShRegOut(9);  -- 0101
      t8(3) := ShRegOut(6);  -- 1101
      t8(4) := ShRegOut(13); -- 0011
      t8(5) := ShRegOut(2);  -- 1011
      t8(6) := ShRegOut(10); -- 0111
      t8(7) := ShRegOut(5);  -- 1111
    end if;

    if (cnt5(1)='0') then
      t4 := t8(3 downto 0);
    else
      t4 := t8(7 downto 4);
    end if;

    if (cnt5(2)='0') then
      t2 := t4(1 downto 0);
    else
      t2 := t4(3 downto 2);
    end if;

    if (cnt5(3)='0') then
      SDout <= t2(0);
    else
      SDout <= t2(1);
    end if;

  end process;

end ActelProAsic3;
